<?php

/**
 * @file
 * Provides general pages and UI functions for the PIFT module.
 */


/**
 * Form callback for per-project DrupalCI configuration.
 */
function pift_ci_project_form(stdClass $project_node) {
  $pift_ci_project = pift_ci_project_info($project_node->nid);

  // Find testable branch releases.

  $testable_branch_releases = [];
  foreach (project_release_query_releases($project_node->nid) as $release) {
    $release_wrapper = entity_metadata_wrapper('node', $release->nid);
    // Lets only display testing configurations for releases that are a. branches. and b. not obsolete.
    if ($release_wrapper->field_release_build_type->value() === 'dynamic' && $release_wrapper->field_release_category->value() != 'obsolete') {
      $testable_branch_releases[$release->nid] = $release;
    }
  }

  // If there are no testable_branch_releases, stop early.
  $output = [];
  if (empty($testable_branch_releases)) {
    $output['message'] = [
      '#prefix' => '<p>',
      '#markup' => t('Please create a dev release to enable automated testing.'),
      '#suffix' => '</p>',
    ];
    return $output;
  }

  if (pift_pages_project_testing_settings_access($project_node)) {
    $output['delete_all'] = [
      '#prefix' => '<p>',
      '#markup' => t('When GitLab CI is working, please <a href="!delete-all">delete all DrupalCI testing schedules</a>', [
        '!delete-all' => url('node/' . $project_node->nid . '/qa/delete_all', ['query' => ['destination' => 'node/' . $project_node->nid]]),
      ]),
      '#suffix' => '</p>',
    ];
  }

  $pift_ci_project_per_release = [];
  foreach ($pift_ci_project as $environment => $ci_releases) {
    foreach (array_keys($ci_releases) as $release_nid) {
      $pift_ci_project_per_release[$release_nid][] = $environment;
    }
  }
  $messages = [
    PIFT_CI_TEST_BRANCH_WEEKLY => t('tested weekly'),
    PIFT_CI_TEST_BRANCH_DAILY => t('tested daily'),
    PIFT_CI_TEST_BRANCH_COMMIT => t('tested on commit'),
    PIFT_CI_TEST_ISSUE => t('tested on commit, issue testing default'),
    PIFT_CI_TEST_DISABLED => t('disabled, testing configuration no longer available'),
  ];
  foreach (array_keys($testable_branch_releases) as $release_branch_nid) {
    $job_ids = [];

    if (isset($pift_ci_project_per_release[$release_branch_nid])) {
      foreach ($pift_ci_project_per_release[$release_branch_nid] as $environment) {
        foreach ($pift_ci_project[$environment][$release_branch_nid] as $core_branch => $trigger) {
          // Latest complete test.
          $query = new EntityFieldQuery();
          $query->entityCondition('entity_type', 'pift_ci_job')
            ->propertyCondition('target_type', PIFT_CI_TYPE_BRANCH)
            ->propertyCondition('release_nid', $release_branch_nid)
            ->propertyCondition('environment', $environment)
            ->propertyCondition('status', [PIFT_CI_STATUS_QUEUE, PIFT_CI_STATUS_SENT, PIFT_CI_STATUS_RUNNING], 'NOT IN')
            ->propertyCondition('custom', FALSE)
            ->propertyOrderBy('job_id', 'DESC')
            ->range(0, 1);
          // $core branch will only be empty if this is a drupal core test.
          if (!empty($core_branch)) {
            // Check current core branch used by release label.
            $query->propertyCondition('core_branch', pift_get_current_core($core_branch, $testable_branch_releases[$release_branch_nid]->field_release_category[LANGUAGE_NONE][0]['value']));

          }
          $result = $query->execute();
          if (isset($result['pift_ci_job'])) {
            $job_ids[array_keys($result['pift_ci_job'])[0]] = ['trigger' => $trigger, 'core_branch' => $core_branch];
          }
          elseif (!empty($core_branch)) {
            // Latest complete test, without specific core branch. For when the
            // value of a release label has recently changed, try to show
            // something.
            $query = new EntityFieldQuery();
            $query->entityCondition('entity_type', 'pift_ci_job')
              ->propertyCondition('target_type', PIFT_CI_TYPE_BRANCH)
              ->propertyCondition('release_nid', $release_branch_nid)
              ->propertyCondition('environment', $environment)
              ->propertyCondition('status', [PIFT_CI_STATUS_QUEUE, PIFT_CI_STATUS_SENT, PIFT_CI_STATUS_RUNNING], 'NOT IN')
              ->propertyCondition('custom', FALSE)
              ->propertyOrderBy('job_id', 'DESC')
              ->range(0, 1);
            $result = $query->execute();
            if (isset($result['pift_ci_job'])) {
              $job_ids[array_keys($result['pift_ci_job'])[0]] = ['trigger' => $trigger, 'core_branch' => $core_branch];
            }
          }
        }
      }

      // Queued tests.
      $query = new EntityFieldQuery();
      $query->entityCondition('entity_type', 'pift_ci_job')
        ->propertyCondition('target_type', PIFT_CI_TYPE_BRANCH)
        ->propertyCondition('release_nid', $release_branch_nid)
        ->propertyCondition('environment', $pift_ci_project_per_release[$release_branch_nid])
        ->propertyCondition('status', [PIFT_CI_STATUS_QUEUE, PIFT_CI_STATUS_SENT, PIFT_CI_STATUS_RUNNING])
        ->propertyCondition('custom', FALSE);
      $result = $query->execute();
      if (isset($result['pift_ci_job'])) {
        $job_ids += array_flip(array_keys($result['pift_ci_job']));
      }
    }

    // Recent custom tests.
    $query = new EntityFieldQuery();
    $query->entityCondition('entity_type', 'pift_ci_job')
      ->propertyCondition('target_type', PIFT_CI_TYPE_BRANCH)
      ->propertyCondition('release_nid', $release_branch_nid)
      ->propertyCondition('custom', TRUE)
      ->propertyCondition('created', REQUEST_TIME - PIFT_CUSTOM_TIME, '>')
      ->propertyOrderBy('job_id', 'DESC');
    $result = $query->execute();
    if (isset($result['pift_ci_job'])) {
      $job_ids += array_fill_keys(array_keys($result['pift_ci_job']), ['trigger' => 'custom']);
    }

    // Render the tests.
    $output[$release_branch_nid]['#prefix'] = '<h3>' . check_plain($testable_branch_releases[$release_branch_nid]->field_release_version[LANGUAGE_NONE][0]['value']) . '</h3>';
    if (empty($job_ids)) {
      $output[$release_branch_nid]['tests'] = pift_ci_render_tests([]);
      $output[$release_branch_nid]['tests'][] = ['#markup' => t('No test results.')];
    }
    else {
      $jobs = entity_load('pift_ci_job', array_keys($job_ids));
      $output[$release_branch_nid]['tests'] = pift_ci_render_tests($jobs);
      $messages_displayed = [];
      $custom_jobs_displayed = [];
      foreach ($jobs as $job_id => $job) {
        if ($job_ids[$job->job_id]['trigger'] === 'custom') {
          // Do not show more than one custom job with the same environment,
          // release node ID, and core branch.
          if (isset($custom_jobs_displayed[$job->environment][$release_branch_nid][$job->core_branch])) {
            unset($output[$release_branch_nid]['tests'][$job_id]);
          }
          else {
            $custom_jobs_displayed[$job->environment][$release_branch_nid][$job->core_branch] = TRUE;
          }
        }
        elseif (isset($messages[$job_ids[$job->job_id]['trigger']]) && !isset($messages_displayed[$job->environment][$release_branch_nid][$job_ids[$job->job_id]['core_branch']][$job_ids[$job->job_id]['trigger']])) {
          $message = $messages[$job_ids[$job->job_id]['trigger']];
          $messages_displayed[$job->environment][$release_branch_nid][$job_ids[$job->job_id]['core_branch']][$job_ids[$job->job_id]['trigger']] = TRUE;
          if (pift_pages_project_testing_settings_access($project_node)) {
            $query = drupal_get_destination();
            $query['environment'] = $job->environment;
            if ($job_ids[$job->job_id]['trigger'] === PIFT_CI_TEST_DISABLED) {
              // Get the original non-disabled trigger.
              $query['schedule'] = pift_ci_project_info($project_node->nid, TRUE)[$job->environment][$release_branch_nid][$job_ids[$job->job_id]['core_branch']];
            }
            else {
              $query['schedule'] = $job_ids[$job->job_id]['trigger'];
            }
            $query['core_branch'] = $job_ids[$job->job_id]['core_branch'];
            $message .= ', ' . l(t('remove'), 'node/' . $project_node->nid . '/qa/add/' . $release_branch_nid, ['query' => $query]);
          }
          $output[$release_branch_nid]['tests'][$job_id]['#suffix'] = ' ' . $message . $output[$release_branch_nid]['tests'][$job_id]['#suffix'];
        }
      }
    }
  }

  return $output;
}

/**
 * Form callback. Delete tests for all branches.
 */
function pift_pages_project_testing_delete_all($form, &$form_state, stdClass $project) {
  drupal_set_title(t('Delete all DrupalCI schedules for @project', ['@project' => $project->title]));
  $form['intro'] = [
    '#prefix' => '<p>',
    '#markup' => t('Are you sure that you want to delete all schedules? <b>This operation cannot be undone.</b>'),
    '#suffix' => '</p>',
  ];
  $form['link_back'] = [
    '#markup' => l(t('Take me back'), 'node/' . $project->nid . '/qa'). ' '
  ];
  $form['submit'] = [
    '#type' => 'submit',
    '#value' => t('Delete all'),
  ];
  $form_state['#pift_delete_all'] = TRUE;

  return $form;
}

/**
 * Form submit callback for pift_pages_project_testing_delete_all().
 */
function pift_pages_project_testing_delete_all_submit($form, &$form_state) {
  if (!empty($form_state['#pift_delete_all'])) {
    db_delete('pift_ci_project')
      ->condition('nid', $form_state['build_info']['args'][0]->nid)
      ->execute();
    drupal_set_message(t('All the configured schedules were deleted.'));
  }
}

/**
 * Form callback. Retesting for branches.
 */
function pift_pages_project_testing_add($form, &$form_state, stdClass $project, stdClass $release) {
  $release_wrapper = entity_metadata_wrapper('node', $release);

  drupal_set_message(t('DrupalCI testing is deprecated. You cannot add new tests or edit existing ones.'), 'warning');

  // Check that this is a valid release.
  if (!project_release_node_is_release($release)) {
    return MENU_NOT_FOUND;
  }
  if (project_release_get_release_project_nid($release) != $project->nid) {
    return MENU_NOT_FOUND;
  }
  if ($release_wrapper->field_release_build_type->value() !== 'dynamic') {
    return MENU_NOT_FOUND;
  }

  $environments = variable_get('pift_ci_environments');
  $pift_ci_project = pift_ci_project_info($project->nid);

  // Custom job.
  $form['custom'] = array(
    '#tree' => TRUE,
    '#type' => 'fieldset',
  );
  // Environment.
  $form['custom']['environment'] = array(
    '#title' => t('Environment'),
    '#type' => 'select',
    '#options' => array(),
    '#disabled' => TRUE,
  );
  if (isset($_GET['environment'])) {
    $form['custom']['environment']['#default_value'] = $_GET['environment'];
  }

  $release_wrapper = entity_metadata_wrapper('node', $release);
  $category = $release_wrapper->field_release_category->value();
  foreach ($environments as $key => $environment) {
    $form['custom']['environment']['#options'][$key] = $environment['title'];
    if (!isset($form['custom']['environment']['#default_value'])) {
      $form['custom']['environment']['#default_value'] = $key;
    }
  }
  // Core.
  if ($project->type === 'project_core') {
    $form['custom']['core_branch'] = [
      '#type' => 'value',
      '#value' => '',
    ];
  }
  else {
    $form['custom']['core_branch'] = [
      '#title' => t('Core'),
      '#type' => 'radios',
      '#required' => TRUE,
      '#description' => l(t('About core branch labels'), 'drupalorg/docs/drupal-ci/drupal-core-branch-labels-in-drupalci'),
      '#disabled' => TRUE,
    ];
    $branch_labels = pift_get_core_labels();
    foreach ($branch_labels as $reference => $config) {
      if ($config->category == $category && $config->status) {
        $form['custom']['core_branch']['#options'][$config->machine_label] = $config->label . ', currently ' . $config->reference;
        $form['custom']['core_branch'][$config->machine_label]['#description'] = $config->label_description;
      }
    }
    if (isset($_GET['core_branch'])) {
      if (isset($form['custom']['core_branch']['#options'][$_GET['core_branch']])) {
        $form['custom']['core_branch']['#default_value'] = $_GET['core_branch'];
      }
    }
    else {
      list($form['custom']['core_branch']['#default_value']) = array_keys($form['custom']['core_branch']['#options']);
    }
  }
  // Schedule.
  $form['custom']['schedule'] = [
    '#title' => t('Schedule'),
    '#type' => 'radios',
    '#options' => [
      '' => t('Run once'),
      PIFT_CI_TEST_BRANCH_WEEKLY => t('Run weekly'),
      PIFT_CI_TEST_BRANCH_DAILY => t('Run daily'),
      PIFT_CI_TEST_BRANCH_COMMIT => t('Run on commit'),
      PIFT_CI_TEST_ISSUE => t('Run on commit and for issues'),
    ],
    '#default_value' => isset($_GET['schedule']) ? $_GET['schedule'] : '',
    '#disabled' => TRUE,
  ];
  foreach ($pift_ci_project as $environment => $project_info) {
    if (isset($project_info[$release->nid]) && in_array(PIFT_CI_TEST_ISSUE, $project_info[$release->nid])) {
      $form['custom']['schedule']['#options'][PIFT_CI_TEST_ISSUE] .= ', ' . t('replacing %environment', ['%environment' => $environments[$environment]['title']]);
      break;
    }
  }

  if (isset($_GET['environment']) && isset($_GET['core_branch']) && isset($_GET['schedule'])) {
    $form_state['#pift_updating'] = TRUE;
    $form['delete'] = [
      '#type' => 'submit',
      '#value' => t('Remove'),
    ];
  }

  return $form;
}

/**
 * Form submit callback for pift_pages_project_testing_add().
 */
function pift_pages_project_testing_add_submit($form, &$form_state) {
  // If updating, clear out old configuration.
  if (!empty($form_state['#pift_updating'])) {
    db_delete('pift_ci_project')
      ->condition('environment', $_GET['environment'])
      ->condition('core_branch', $_GET['core_branch'])
      ->condition('testing', $_GET['schedule'])
      ->condition('nid', $form_state['build_info']['args'][0]->nid)
      ->condition('release_nid', $form_state['build_info']['args'][1]->nid)
      ->execute();
  }

  if ($form_state['triggering_element']['#id'] === 'edit-delete') {
    drupal_set_message('Removed testing schedule.');
  }
}

/**
 * Form for “Add test / retest” pages.
 */
function pift_pages_issue_testing($form, &$form_state, stdClass $node, $target_type, $tests_key) {
  $project = node_load($node->field_project[LANGUAGE_NONE][0]['target_id']);
  project_issue_set_breadcrumb($node, $project);
  $breadcrumb = drupal_get_breadcrumb();
  $breadcrumb[] = l($node->title, 'node/' . $node->nid);
  drupal_set_breadcrumb($breadcrumb);

  $tests = pift_load_tests($node, $target_type);
  $previous_tests = [];
  if ($target_type === PIFT_CI_TYPE_MERGE_REQUEST) {
    $group_by = 'commit_id';
    preg_match(PIFT_MERGE_REQUEST_ARG_REGEXP, $tests_key, $match);
    if (!isset($tests[$match['branch']])) {
      // This should not normally be reachable. We use an existing test to
      // determine which release to use for testing configuration.
      drupal_set_message(t('There are no tests for this merge request yet. The first test must be triggered with a git push.'), 'warning');
      return $form;
    }
    foreach ($tests[$match['branch']] as $previous_test) {
      if (!isset($release)) {
        $release = node_load($previous_test->release_nid);
      }
      if (!isset($form_state['pift_merge_target'])) {
        $form_state['pift_merge_target'] = $previous_test->merge_target;
      }
      if ($previous_test->commit_id === $match['commit']) {
        $previous_tests[] = $previous_test;
      }
    }
  }
  else {
    $group_by = 'release_nid';
    $release = pift_ci_find_dynamic_release($node);
    if (isset($tests[$tests_key])) {
      $previous_tests = $tests[$tests_key];
    }
  }
  $form_state['pift_default_release_wrapper'] = entity_metadata_wrapper('node', $release);

  $form['tests_heading'] = [];
  $form['tests'] = pift_ci_display_tests($node, $previous_tests, $group_by);

  $form['add_test'] = array(
    '#type' => 'checkboxes',
    '#options' => array(),
  );

  // Custom job.
  $form['custom'] = array(
    '#tree' => TRUE,
    '#type' => 'fieldset',
  );
  if ($target_type === PIFT_CI_TYPE_FILE) {
    $form['custom']['release'] = [
      '#title' => t('Release'),
      '#type' => 'select',
      '#options' => [],
      '#default_value' => $release->nid,
      '#states' => [
        'enabled' => [
          'input[name="add_test[custom]"]' => ['checked' => TRUE],
        ],
      ],
    ];
  }
  $form['custom']['environment'] = array(
    '#title' => t('Environment'),
    '#type' => 'select',
    '#options' => array(),
    '#states' => array(
      'enabled' => array(
        'input[name="add_test[custom]"]' => array('checked' => TRUE),
      ),
    ),
  );

  $form_state['existing_tests'] = [];
  $project_info = pift_ci_project_info($project->nid);
  $environments = variable_get('pift_ci_environments');
  foreach ($project_info as $environment => $environment_info) {
    if (!isset($environment_info[$release->nid])) {
      continue;
    }

    foreach ($environment_info[$release->nid] as $core_branch => $testing) {
      $core_branch = pift_get_current_core($core_branch, $form_state['pift_default_release_wrapper']->field_release_category->value());

      // Check what tests are already existing.
      $in_progress = FALSE;
      $key = $environment . '--' . $core_branch;
      if (!empty($previous_tests)) {
        foreach ($previous_tests as $job) {
          if (empty($form['tests_heading'])) {
            $form['tests_heading']['#markup'] = '<p>' . t('Results for @target:', [
              '@target' => $job->getTargetTitle(),
            ]) . '</p>';
          }
          if ($environment === $job->environment && $core_branch === $job->core_branch) {
            if ($job->isRunning()) {
              $in_progress = TRUE;
            }
            else {
              $form_state['existing_tests'][$key] = $job;
            }
          }
        }
      }
      if (!$in_progress && $testing !== PIFT_CI_TEST_DISABLED) {
        // Add retest or test.
        $form['add_test']['#options'][$key] = t(isset($form_state['existing_tests'][$key]) ? 'Retest @test' : 'Test @test', ['@test' => $environments[$environment]['title']]);
        if (!empty($core_branch)) {
          $form['add_test']['#options'][$key] .= ', D' . preg_replace('/\.x$/', '', $core_branch);
        }
      }
    }
  }

  // Populate custom options.
  foreach (variable_get('pift_ci_environments') as $key => $environment) {
    $form['custom']['environment']['#options'][$key] = $environment['title'];
    if (isset($project_info[$key])) {
      if (isset($project_info[$key][$release->nid]) && in_array(PIFT_CI_TEST_ISSUE, $project_info[$key][$release->nid])) {
        $form['custom']['environment']['#default_value'] = $key;
      }
      if ($target_type === PIFT_CI_TYPE_FILE) {
        foreach (array_keys($project_info[$key]) as $release_nid) {
          $form['custom']['release']['#options'][$release_nid] = TRUE;
        }
      }
    }
  }

  // Add release titles.
  $categories = [];
  switch ($target_type) {
    case PIFT_CI_TYPE_FILE:
      foreach (node_load_multiple(array_keys($form['custom']['release']['#options'])) as $node) {
        $form['custom']['release']['#options'][$node->nid] = check_plain($node->title);
        $categories[$node->field_release_category[LANGUAGE_NONE][0]['value']][] = $node->nid;
      }
      break;

    case PIFT_CI_TYPE_MERGE_REQUEST:
      $categories[$form_state['pift_default_release_wrapper']->field_release_category->value()][] = $release->nid;
      break;
  }

  // For non-core tests, add a Core parameter.
  if ($project->type !== 'project_core') {
    // The release category of core needs to match the release category of the
    // release being tested with. Make a core select element for each release
    // category, and make visible based on the currently-selected release.
    foreach ($categories as $category => $release_nids) {
      $form['custom']['core_branch-' . $category] = [
        '#title' => t('Core'),
        '#type' => 'select',
        '#options' => [],
        '#states' => [
          'enabled' => [
            'input[name="add_test[custom]"]' => ['checked' => TRUE],
          ],
          'visible' => [],
        ],
      ];
      if ($target_type === PIFT_CI_TYPE_FILE) {
        foreach ($release_nids as $release_nid) {
          $form['custom']['core_branch-' . $category]['#states']['visible']['select[name="custom[release]"]'][] = ['value' => $release_nid];
        }
      }
    }
    $branch_labels = pift_get_core_labels();
    foreach ($branch_labels as $reference => $config) {
      if ($config->status) {
        $form['custom']['core_branch-' . $config->category]['#options'][$config->reference] = 'Drupal ' . $config->reference . ' : ' . $config->label;
      }
    }
  }

  $form['add_test']['#options']['custom'] = t('Custom parameters');

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Queue'),
  );

  return $form;
}

/**
 * @return mixed
 */
function pift_get_core_labels() {
  $query = db_select('pift_ci_label_map', 'l');
  $query->join('field_data_field_release_category', 'fdfrc', 'fdfrc.entity_id = l.release_nid');
  $query->join('field_data_field_release_build_type', 'fdfrbt', 'fdfrbt.entity_id = l.release_nid');
  $query->fields('l', [
    'machine_label',
    'label',
    'label_description',
    'reference',
    'status',
    'release_nid',
  ]);
  $query->addField('fdfrc', 'field_release_category_value', 'category');
  $query->orderBy('fdfrbt.field_release_build_type_value', 'ASC');
  $query->orderBy('l.release_nid', 'ASC');
  $branch_label = $query->execute()->fetchAllAssoc('reference');
  return $branch_label;
}

/**
 * Form validation callback for pift_pages_issue_testing().
 */
function pift_pages_issue_testing_validate($form, &$form_state) {
  if (count(array_filter($form_state['values']['add_test'])) == 0) {
    form_set_error('add_test', t('No tests selected!'));
  }
}

/**
 * Form submit callback for pift_pages_issue_testing().
 */
function pift_pages_issue_testing_submit($form, &$form_state) {
  foreach (array_keys(array_filter($form_state['values']['add_test'])) as $key) {
    $job_template = [];

    switch ($form_state['build_info']['args'][1]) {
      case PIFT_CI_TYPE_FILE:
        $job_template['file_id'] = $form_state['build_info']['args'][2];
        if ($key === 'custom') {
          $job_template['release_nid'] = $form_state['values']['custom']['release'];
        }
        elseif ($release_nid = db_query('SELECT release_nid FROM {pift_file} WHERE file_id = :fid', [':fid' => $form_state['build_info']['args'][2]])->fetchField()) {
          $job_template['release_nid'] = $release_nid;
        }
        else {
          $job_template['release_nid'] = pift_ci_find_dynamic_release($form_state['build_info']['args'][0])->nid;
        }
        break;

      case PIFT_CI_TYPE_MERGE_REQUEST:
        preg_match(PIFT_MERGE_REQUEST_ARG_REGEXP, $form_state['build_info']['args'][2], $match);
        $job_template['issue_branch'] = $match['branch'];
        $job_template['commit_id'] = $match['commit'];
        $job_template['release_nid'] = $form_state['pift_default_release_wrapper']->getIdentifier();
        $job_template['merge_target'] = $form_state['pift_merge_target'];
        break;
    }

    if ($key === 'custom') {
      $job = new PiftCiJob([
        'reason' => '!name added custom test.',
        'environment' => $form_state['values']['custom']['environment'],
        'target_type' => $form_state['build_info']['args'][1],
        'issue_nid' => $form_state['build_info']['args'][0]->nid,
        'custom' => TRUE,
      ] + $job_template);
      if (isset($form_state['values']['custom']['core_branch-' . $job->getReleaseCategory()])) {
        $job->core_branch = $form_state['values']['custom']['core_branch-' . $job->getReleaseCategory()];
      }
      $job->save();
      drupal_set_message(t('%target has been queued for custom testing.', [
        '%target' => $job->getTargetTitle(),
      ]));
    }
    else {
      list($environment, $core_branch) = explode('--', $key);
      $job = new PiftCiJob([
        'reason' => '!name added test.',
        'environment' => $environment,
        'core_branch' => $core_branch,
        'target_type' => $form_state['build_info']['args'][1],
        'issue_nid' => $form_state['build_info']['args'][0]->nid,
      ] + $job_template);
      if (isset($form_state['existing_tests'][$key])) {
        $job->reason = '!name added re-test.';
        $job->build_iteration = $form_state['existing_tests'][$key]->build_iteration + 1;
      }
      $job->save();
      drupal_set_message(t('%target has been queued for testing with %environment.', [
        '%target' => $job->getTargetTitle(),
        '%environment' => variable_get('pift_ci_environments')[$environment]['title'],
      ]));
    }
  }

  $form_state['redirect'] = 'node/' . $form_state['build_info']['args'][0]->nid;
}

/**
 * Menu callback, page for a PIFT CI job.
 */
function pift_ci_job_page(PiftCiJob $job) {
  // Load project* nodes and set breadcrumbs.
  $release_node = node_load($job->release_nid);
  $release_wrapper = entity_metadata_wrapper('node', $release_node);
  if (!empty($job->issue_nid)) {
    $issue_node = node_load($job->issue_nid);
    $project_node = node_load($issue_node->field_project[LANGUAGE_NONE][0]['target_id']);
    project_issue_set_breadcrumb($issue_node, $project_node);
    $breadcrumb = drupal_get_breadcrumb();
    $breadcrumb[] = l($issue_node->title, 'node/' . $issue_node->nid);
    drupal_set_breadcrumb($breadcrumb);
  }
  else {
    $project_node = node_load(project_release_get_release_project_nid($release_node));
    project_project_set_breadcrumb($project_node, TRUE);
    $breadcrumb = drupal_get_breadcrumb();
    $breadcrumb[] = l(t('Automated Testing'), 'node/' . $project_node->nid . '/qa');
    drupal_set_breadcrumb($breadcrumb);
  }
  drupal_set_title($job->getTitle(), PASS_THROUGH);

  $output = array(
    'meta' => array(
      '#prefix' => '<p class="submitted">',
      'created' => array(
        '#markup' => t('Created @created.', array('@created' => format_date($job->created))) . ' ',
      ),
      'reason' => array(
        '#markup' => t($job->reason, array(
          '!name' => theme('username', array('account' => user_load($job->uid))),
          '!bypass_name' => theme('username', array('account' => user_load($job->bypass_uid))),
          '@commit_id' => $job->commit_id,
          '@issue_branch' => $job->issue_branch,
        )) . ' ',
      ),
      'updated' => array(
        '#markup' => t('Updated @updated.', array('@updated' => format_date($job->updated))),
      ),
      '#suffix' => '</p>',
    ),
    'results_link' => array(
      '#prefix' => '<p>' . $job->message . ' ',
      '#suffix' => '</p>',
    ),
    'build_details' => array(),
    'cancel' => array(),
    'notification' => array(),
    'results' => array(),
    'checkstyle' => [],
    'branch' => array(),
    'history' => array(
      '#prefix' => '<h2>' . t('History') . '</h2>',
      '#theme' => 'table',
      '#header' => array(t('Updated'), t('Result')),
      '#empty' => t('No previous results.'),
      '#rows' => array(),
      '#attributes' => ['class' => ['pift-history']],
    ),
  );

  $output['results_link']['#markup'] = '<div class="warning">' . t('DrupalCI console output and artifacts are no longer available from dispatcher.drupalci.org. <a href="@link">Use GitLab CI instead</a>', [
    '@link' => url('docs/develop/git/using-gitlab-to-contribute-to-drupal/gitlab-ci'),
  ]) . '</div>';

  if ($job->status == PIFT_CI_STATUS_ERROR && (!empty($job->build_details))) {
    $output['build_details']['#markup'] = '<pre>' . check_plain(truncate_utf8(trim($job->build_details), 1000000, FALSE, TRUE)) . '</pre>';
  }

  // Cancel form.
  if ($job->isRunning() && $job->editAccess()) {
    $output['cancel'] = drupal_get_form('pift_ci_cancel_form', $job);
  }

  // Notification on results.
  if (user_is_logged_in()) {
    // Show existing notifications.
    $notify_on = array();
    foreach ($job->getNotifications($GLOBALS['user']) as $notification) {
      $notify_on[$notification->notify_on] = TRUE;
    }
    if (isset($notify_on[PIFT_CI_SUBSCRIBE_ALL])) {
      $output['notification']['#markup'] = '<p>' . t('✓ You will be notified when this test completes.') . ' ' . l('edit', 'user/' . $GLOBALS['user']->uid . '/project-issue') . '</p>';
    }
    elseif (isset($notify_on[PIFT_CI_SUBSCRIBE_NONPASS])) {
      $output['notification']['#markup'] = '<p>' . t('✓ You will be notified if this test does not pass.') . ' ' . l('edit', 'user/' . $GLOBALS['user']->uid . '/project-issue') . '</p>';
    }
    else {
      // No existing notification.
      $output['notification'] = drupal_get_form('pift_ci_notification_form', $job);
    }
  }

  // Related branch results.
  if (in_array($job->target_type, [PIFT_CI_TYPE_FILE, PIFT_CI_TYPE_MERGE_REQUEST])) {
    if ($branch_job = $job->getCurrentBranchJob()) {
      // If the branch is blocking a test from getting out of the queue.
      if (user_access('pift re-test files') && $job->status === PIFT_CI_STATUS_WAITING) {
        $output['branch']['bypass'] = drupal_get_form('pift_ci_bypass_branch_form', $job);
      }

      $output['branch']['current'] = pift_ci_render_tests([$branch_job]);
      $output['branch']['current'][$branch_job->job_id]['#suffix'] = ' ' . t('most recent, @date', array('@date' => format_date($branch_job->updated))) . $output['branch']['current'][$branch_job->job_id]['#suffix'];
    }
    if (($last_branch_job = $job->getLastBranchJob()) && $last_branch_job->job_id !== $branch_job->job_id) {
      $output['branch']['last'] = pift_ci_render_tests([$last_branch_job]);
      $output['branch']['last'][$last_branch_job->job_id]['#suffix'] = ' ' . t('as of issue test creation, @date', array('@date' => format_date($last_branch_job->updated))) . $output['branch']['last'][$last_branch_job->job_id]['#suffix'];
    }
    if (!empty($output['branch'])) {
      $output['branch']['#prefix'] = '<h2>' . t('@branch branch result', array('@branch' => $release_node->title)) . '</h2>';
    }
  }

  if ($job->status === PIFT_CI_STATUS_SWEPT) {
    $output['results'] = ['#markup' => '<p><em>' . t('Outdated results are no longer available.') . '</em></p>'];
  }
  else {
    // Results line items.
    $skipped_count = 0;
    foreach ($job->getResultArray() as $suite => $suite_results) {
      // If more than 1⁄3 of the memory is used at this point, there won't be
      // enough to show everything.
      if (!drupal_check_memory_limit(memory_get_usage() * 3)) {
        $skipped_count += array_sum(array_map('count', $suite_results));
        continue;
      }

      // Suite.
      $output['results']['#rows'][$suite] = array(
        'data' => array(
          PIFT_CI_RESULT_PASS => array(
            'data' => NULL,
            'class' => array('pift-ci-result-column'),
          ),
          PIFT_CI_RESULT_FAIL => array(
            'data' => NULL,
            'class' => array('pift-ci-result-column'),
          ),
          check_plain($suite),
        ),
        'class' => array('pift-ci-suite'),
        'no_striping' => TRUE,
      );

      foreach ($suite_results as $class => $class_results) {
        // Class.
        $output['results']['#rows'][$class] = array(
          'data' => array(
            PIFT_CI_RESULT_PASS => array(
              'data' => NULL,
              'class' => array('pift-ci-result-column'),
            ),
            PIFT_CI_RESULT_FAIL => array(
              'data' => NULL,
              'class' => array('pift-ci-result-column'),
            ),
            check_plain($class),
          ),
          'class' => array('pift-ci-class'),
          'no_striping' => TRUE,
        );

        foreach ($class_results as $job_result) {
          // Increment pass and fail sums.
          $output['results']['#rows'][$job_result->suite]['data'][$job_result->result]['data'] += 1;
          $output['results']['#rows'][$job_result->class]['data'][$job_result->result]['data'] += 1;

          // Test output.
          $text = '';
          if (!empty($job_result->output)) {
            $text .= '<pre>' . check_plain(truncate_utf8(trim($job_result->output), 1000000, FALSE, TRUE)) . '</pre>';
          }
          if (!empty($job_result->error)) {
            $text .= '<pre>' . check_plain(truncate_utf8(trim($job_result->error), 1000000, FALSE, TRUE)) . '</pre>';
          }

          $output['results']['#rows'][] = array(
            'data' => array(
              array(
                'data' => $job_result->result === PIFT_CI_RESULT_PASS ? '<span class="pift-ci-icon">✓</span>' : '',
                'class' => array('pift-ci-result-column'),
              ),
              array(
                'data' => $job_result->result === PIFT_CI_RESULT_FAIL ? '<span class="pift-ci-icon">✗</span>' : '',
                'class' => array('pift-ci-result-column'),
              ),
              array(
                'data' => empty($text) ? '<span class="pift-ci-no-output">-</span> ' . check_plain($job_result->test) : theme('ctools_collapsible', array(
                  'handle' => check_plain($job_result->test),
                  'content' => $text,
                  'collapsed' => $job_result->result === PIFT_CI_RESULT_PASS,
                )),
              ),
            ),
            'class' => array('pift-ci-' . $job_result->result),
            'no_striping' => TRUE,
          );
        }
      }
    }
    if ($skipped_count > 0) {
      $output['results']['#rows'][] = [
        'data' => [
          [
            'data' => '<em>' . t('Too much test output, @skipped_count tests skipped.', ['@skipped_count' => number_format($skipped_count)]) . '</em>',
            'colspan' => 3,
          ],
        ],
        'class' => ['pift-ci-message'],
        'no_striping' => TRUE,
      ];
    }
    if (isset($output['results']['#rows'])) {
      foreach ($output['results']['#rows'] as &$result_row) {
        if ($result_row['class'][0] === 'pift-ci-suite') {
          if (empty($result_row['data']['fail']['data'])) {
            $result_row['class'][] = 'pift-ci-all-pass';
          }
        }
        elseif ($result_row['class'][0] === 'pift-ci-class') {
          if ($all_pass = empty($result_row['data']['fail']['data'])) {
            $result_row['class'][] = 'pift-ci-all-pass';
          }
        }
        else {
          if ($all_pass) {
            $result_row['class'][] = 'pift-ci-all-pass';
          }
        }
      }
      // Add table information.
      $output['results']['#prefix'] = '<div class="pift-ci-toggle">' . t('Show: <a href="#" class="active">Failing classes</a> <a href="#">All</a>') . '</div>';
      $output['results']['#theme'] = 'table';
      $output['results']['#attributes'] = array(
        'class' => array('pift-ci-results', 'pift-ci-hide-all-pass'),
      );
      if ($job->result === PIFT_CI_RESULT_PASS) {
        array_unshift($output['results']['#rows'], array(
          'data' => array(
            array(
              'data' => '<em>' . t('All classes passed!') . '</em>',
              'colspan' => 3,
            ),
          ),
          'no_striping' => TRUE,
        ));
      }
      $output['results']['#header'] = array(
        array(
          'data' => t('✓'),
          'class' => array('pift-ci-result-column'),
        ),
        array(
          'data' => t('✗'),
          'class' => array('pift-ci-result-column'),
        ),
        NULL,
      );
    }
  }

  // Checkstyle results.
  if ($job->target_type === PIFT_CI_TYPE_BRANCH) {
    $files = [];
    $repository = versioncontrol_project_repository_load($project_node->nid);
    $vcs_branch = $repository->loadBranch($release_wrapper->field_release_vcs_label->value());
  }
  $n = 0;
  $last_file = '';
  $file_count = 0;
  if (isset($last_branch_job)) {
    $cs_diff_job = $last_branch_job;
  }
  elseif ($job->target_type === PIFT_CI_TYPE_BRANCH) {
    $query = new EntityFieldQuery();
    $result = $query->entityCondition('entity_type', 'pift_ci_job')
      ->propertyCondition('target_type', $job->target_type)
      ->propertyCondition('environment', $job->environment)
      ->propertyCondition('release_nid', $job->release_nid)
      ->propertyCondition('issue_nid', $job->issue_nid)
      ->propertyCondition('file_id', $job->file_id)
      ->propertyCondition('issue_branch', $job->issue_branch)
      ->propertyCondition('custom', $job->custom)
      ->propertyCondition('job_id', $job->job_id, '<')
      ->propertyOrderBy('job_id', 'DESC')
      ->range(0, 1)
      ->execute();
    if (isset($result['pift_ci_job'])) {
      $cs_diff_job = entity_load_single('pift_ci_job', array_keys($result['pift_ci_job'])[0]);
    }
  }
  $result = db_query('SELECT file, line, message FROM {pift_ci_job_checkstyle} WHERE job_id = :job_id ORDER BY file, line', [':job_id' => $job->job_id]);
  foreach ($result as $checkstyle_result) {
    if ($checkstyle_result->file !== $last_file) {
      if (isset($cs_diff_job)) {
        _pift_ci_job_page_checkstyle_diff($output, $last_file, $cs_diff_job, $file_count);
      }

      $last_file = $checkstyle_result->file;
      $file_count = 0;
      $output['checkstyle']['#rows'][$checkstyle_result->file] = [
        'data' => [
          [
            'data' => '<strong>' . check_plain($last_file) . '</strong>',
            'class' => ['pift-ci-file'],
            'id' => drupal_clean_css_identifier('cs-' . $last_file),
            'colspan' => 2,
          ],
        ],
        'no_striping' => TRUE,
      ];
    }
    $file_count += 1;
    $output['checkstyle']['#rows'][$n] = [
      'data' => [
        [
          'data' => $checkstyle_result->line,
          'class' => ['pift-ci-line-column'],
        ],
        check_plain($checkstyle_result->message),
      ],
      'no_striping' => TRUE,
    ];
    if (isset($files)) {
      $files['/' . $checkstyle_result->file][] = $n;
    }
    $n += 1;

    // If more than 1⁄3 of the memory is used at this point, there won't be
    // enough to show everything.
    if (!drupal_check_memory_limit(memory_get_usage() * 3)) {
      $count = 0;
      foreach ($result as $dummy) {
        $count += 1;
      }
      $output['checkstyle']['#rows'][] = array(
        'data' => [
          [
            'data' => '<em>' . t('Too much output, @count messages skipped.', ['@count' => number_format($count)]) . '</em>',
            'colspan' => 2,
          ],
        ],
        'class' => ['pift-ci-message'],
        'no_striping' => TRUE,
      );
    }
  }
  $checkstyle_title = format_plural($job->checkstyle, '1 coding standards message', '@count coding standards messages');
  $branch_diff = 0;
  if (isset($cs_diff_job)) {
    $branch_diff = _pift_ci_job_page_checkstyle_diff($output, $last_file, $cs_diff_job, $file_count);
  }
  if (isset($output['checkstyle']['#rows'])) {
    $output['checkstyle']['#prefix'] = '<h3><a class="pift-checkstyle-toggle">' . $checkstyle_title . '</a></h3>';
    $args = [
      '@diff' => abs($branch_diff),
      '!url' => url('pift-ci-job/' . $cs_diff_job->job_id),
      '@label' => $job->target_type === PIFT_CI_TYPE_BRANCH ? t('previous result') : t('branch result'),
    ];
    if ($branch_diff > 0) {
      $output['checkstyle']['#prefix'] .= '<p class="pift-ci-fail">' . t('<span class="pift-ci-icon">✗</span> @diff more than <a href="!url">@label</a>', $args) . '</p>';
    }
    elseif ($branch_diff < 0) {
      $output['checkstyle']['#prefix'] .= '<p class="pift-ci-pass">' . t('<span class="pift-ci-icon">✓</span> @diff fewer than <a href="!url">@label</a>', $args) . '</p>';
    }
    $output['checkstyle']['#theme'] = 'table';
    $output['checkstyle']['#sticky'] = FALSE;
    $output['checkstyle']['#attributes'] = [
      'class' => ['pift-ci-checkstyle'],
    ];
    // Empty heading to set fixed table layout widths.
    $output['checkstyle']['#header'] = [
      [
        'class' => ['pift-ci-line-column'],
      ],
      NULL,
    ];
    // Use the first line number as a heading.
    $output['checkstyle']['#rows'][0]['data'][0]['data'] = t('line !line', ['!line' => $output['checkstyle']['#rows'][0]['data'][0]['data']]);
  }
  elseif ($job->checkstyle) {
    // There are results which have been removed.
    $output['checkstyle']['#prefix'] = '<h3>' . $checkstyle_title . '</h3>';
    $output['checkstyle']['#suffix'] = '<p><em>' . t('Outdated coding standards messages are not available.') . '</em></p>';
  }

  // Link to previous results.
  $query = new EntityFieldQuery();
  $result = $query->entityCondition('entity_type', 'pift_ci_job')
    ->propertyCondition('target_type', $job->target_type)
    ->propertyCondition('environment', $job->environment)
    ->propertyCondition('release_nid', $job->release_nid)
    ->propertyCondition('issue_nid', $job->issue_nid)
    ->propertyCondition('file_id', $job->file_id)
    ->propertyCondition('issue_branch', $job->issue_branch)
    ->propertyCondition('custom', $job->custom)
    ->propertyOrderBy('job_id', 'DESC')
    ->execute();
  foreach (entity_load('pift_ci_job', array_keys($result['pift_ci_job'])) as $history_job) {
    $history_render = pift_ci_render_tests([$history_job]);
    $row = array(
      'data' => array(
        array('data' => format_date($history_job->updated)),
        array('data' => drupal_render($history_render)),
      ),
      'no_striping' => TRUE,
    );
    if ($history_job->updated === $job->updated) {
      $row['data'][0]['class'][] = 'active';
      $row['data'][1]['class'][] = 'active';
    }
    $output['history']['#rows'][] = $row;
  }

  return $output;
}

/**
 * Append difference with branch job to file heading, if there is one.
 */
function _pift_ci_job_page_checkstyle_diff(&$output, $last_file, $cs_diff_job, $file_count) {
  static $branch_count;
  static $branch_diff = 0;

  if ($branch_count !== FALSE && $last_file !== '') {
    if (!isset($branch_count)) {
      $branch_count = db_query('SELECT file, count(1) FROM {pift_ci_job_checkstyle} WHERE job_id = :job_id GROUP BY file', [':job_id' => $cs_diff_job->job_id])->fetchAllKeyed();
      if (empty($branch_count) && $cs_diff_job->checkstyle != 0) {
        // The branch job’s checkstyle results have been cleaned, there will be
        // no diff.
        $branch_count = FALSE;
        return FALSE;
      }
    }

    $diff = $file_count - (isset($branch_count[$last_file]) ? $branch_count[$last_file] : 0);
    $branch_diff += $diff;
    if ($diff > 0) {
      $output['checkstyle']['#rows'][$last_file]['data'][0]['data'] .= ' ' . l(t('<span class="pift-ci-icon">✗</span> @diff more', ['@diff' => abs($diff)]), 'pift-ci-job/' . $cs_diff_job->job_id, ['html' => TRUE, 'attributes' => ['title' => t('than branch result')], 'fragment' => drupal_clean_css_identifier('cs-' . $last_file)]);
      $output['checkstyle']['#rows'][$last_file]['data'][0]['class'][] = 'pift-ci-fail';
    }
    elseif ($diff < 0) {
      $output['checkstyle']['#rows'][$last_file]['data'][0]['data'] .= ' ' . l(t('<span class="pift-ci-icon">✓</span> @diff fewer', ['@diff' => abs($diff)]), 'pift-ci-job/' . $cs_diff_job->job_id, ['html' => TRUE, 'attributes' => ['title' => t('than branch result')], 'fragment' => drupal_clean_css_identifier('cs-' . $last_file)]);
      $output['checkstyle']['#rows'][$last_file]['data'][0]['class'][] = 'pift-ci-pass';
    }
  }

  return $branch_diff;
}

/**
 * Implements hook_query_TAG_alter().
 */
function pift_query_pift_per_path_alter(\QueryAlterableInterface $query) {
  // We only want one result per-path, and do not really care which one we get.
  $query->groupBy('path');
  $query->orderBy('NULL');
}

/**
 * Form callback. See pift_ci_job_page().
 */
function pift_ci_bypass_branch_form($form, $form_state, $job) {
  return array(
    '#job' => $job,
    'submit' => array(
      '#type' => 'submit',
      '#value' => t('Do not wait for branch to pass'),
      '#suffix' => '<p class="description">' . t('This test will likely have the same failures as the branch, unless they are fixed by this patch.') . '</p>',
    ),
  );
}

/**
 * Form submit callback. See pift_ci_bypass_branch_form().
 */
function pift_ci_bypass_branch_form_submit($form) {
  $form['#job']->status = PIFT_CI_STATUS_QUEUE;
  $form['#job']->message = 'Queueing';
  $form['#job']->bypass_uid = $GLOBALS['user']->uid;
  $form['#job']->reason .= ' !bypass_name bypassed the failing branch.';
  $form['#job']->save();

  drupal_set_message(t('This test will proceed without waiting for the branch to pass.'));
}

/**
 * Form callback. Notification on test results.
 */
function pift_ci_notification_form($form, $form_state, $job) {
  $form['#attributes'] = array('class' => array('container-inline'));
  $form['#job'] = $job;

  $options = array();
  if ($job->isRunning()) {
    $options['test'] = t('for this test');
  }
  //$options['own'] = t('to all tests I request');
  switch ($job->target_type) {
    case PIFT_CI_TYPE_BRANCH:
      $pift_ci_environments = variable_get('pift_ci_environments');
      $options['branch'] = t('to all tests for !branch with !environment', array(
        '!branch' => node_load($job->release_nid)->field_release_version[LANGUAGE_NONE][0]['value'],
        '!environment' => $pift_ci_environments[$job->environment]['title'],
      ));
      break;

    case PIFT_CI_TYPE_FILE:
      //$options['file'] = t('to all tests for @file');
      break;
  }
  if (empty($options)) {
    return;
  }
  $form['type'] = array(
    '#type' => 'select',
    '#title' => t('Add email notification'),
    '#options' => $options,
  );

  $form['notify_on'] = array(
    '#type' => 'select',
    '#title' => t('Test result'),
    '#title_display' => 'none',
    '#options' => array(
      PIFT_CI_SUBSCRIBE_NONPASS => t('if the test does not pass'),
      PIFT_CI_SUBSCRIBE_ALL => t('when the test completes'),
    ),
  );

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Add'),
  );

  return $form;
}

/**
 * Form submit callback for pift_ci_notification_form().
 */
function pift_ci_notification_form_submit($form, $form_state) {
  $fields = array(
    'uid' => $GLOBALS['user']->uid,
    'notify_on' => $form_state['values']['notify_on'],
  );

  switch ($form_state['values']['type']) {
    case 'test':
      $fields['job_id'] = $form['#job']->job_id;
      break;

    case 'branch':
      $fields['release_nid'] = $form['#job']->release_nid;
      $fields['environment'] = $form['#job']->environment;
      break;
  }

  $query = db_insert('pift_ci_notification')
    ->fields($fields)
    ->execute();

  drupal_set_message(t('Notification added.'));
}

/**
 * Form callback. Cancel test.
 */
function pift_ci_cancel_form($form, $form_state, $job) {
  // todo: implement cancelling jobs that have not started running.
  if (empty($job->ci_url)) {
    return;
  }

  $form['#job'] = $job;

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Cancel test'),
  );

  return $form;
}

/**
 * Form submit callback for pift_ci_cancel_form().
 */
function pift_ci_cancel_form_submit($form, $form_state) {
  if ($form['#job']->stopDispatcherJob()) {
    // Under-promise, job might have completed in the meantime, for exmaple.
    drupal_set_message(t('Cancellation requested. This page will update in a couple minutes if successful.'));
  }
  else {
    drupal_set_message(t('Unable to request cancellation.'));
  }
}

/**
 * Menu callback, push update to a job.
 */
function pift_ci_job_ping_page() {
  // Parse the JSON payload.
  $data = json_decode(file_get_contents('php://input'));
  if (empty($data)) {
    watchdog('pift_ci', 'Can’t parse pushed job data.', [], WATCHDOG_WARNING);
    return '';
  }
  $ci_url = variable_get('pift_ci_dispatcher') . '/' . rtrim($data->build->url, '/');

  // Get the parameters of the build.
  try {
    $job_response = pift_get_jenkins_client()->get($ci_url . '/api/json');
  }
  catch (\GuzzleHttp\Exception\RequestException $e) {
    watchdog('pift_ci', 'CI result details fetch failure: @error; @data', ['@error' => $e->getMessage(), '@data' => \GuzzleHttp\Psr7\str($e->getResponse())], WATCHDOG_WARNING);
    return '';
  }
  $job_data = json_decode($job_response->getBody()->getContents());
  foreach ($job_data->actions as $action) {
    if (isset($action->parameters)) {
      foreach ($action->parameters as $parameter) {
        if ($parameter->name === 'Drupal_JobID' && !empty($parameter->value) && strpos($parameter->value, $GLOBALS['base_url'] . ':') === 0) {
          if ($job = PiftCiJob::load((int) str_replace($GLOBALS['base_url'] . ':', '', $parameter->value))) {
            preg_match('#^(?<base>.*)/(?<number>\d+)$#', $ci_url, $match);
            preg_match('#^(?<base>.*)/(?<number>\d+)$#', $job->ci_url, $job_match);
            if (empty($job->ci_url) || ($job->ci_url !== $ci_url && !($job_match['base'] === $match['base'] && $job_match['number'] > $match['number']))) {
              // Job is not already updated and does not have a more recent build number.
              $job->status = PIFT_CI_STATUS_RUNNING;
              $job->message = t('Running');
              $job->ci_url = $ci_url;
              $job->save();
            }
            if (!$job_data->building) {
              // Finished building, get result.
              $job->requestResult();
            }
          }
          return '';
        }
      }
    }
  }
  return '';
}
